/* rcsmtp.c
**
** SMTP interpreter which should run on server's side
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "vxerror.h"
#include "tools.h"


#include "rcsmtp.h"

typedef int (*SMTPcmd_t) (FILE *f, char *arg, void *out);
/* which commands does SMTP have?
 * HELO <SP> <domain> <CRLF>
 * MAIL <SP> FROM:<reverse-path> <CRLF>
 * RCPT <SP> TO:<forward-path> <CRLF>
 * DATA <CRLF>
 * RSET <CRLF>
 * SEND <SP> FROM:<reverse-path> <CRLF>    -- send (to a terminal) isof MAIL
 * SOML <SP> FROM:<reverse-path> <CRLF>    -- Send Or MaiL
 * SAML <SP> FROM:<reverse-path> <CRLF>    -- Send And MaiL
 * VRFY <username> <CRLF>
 *      250 real name <email-address>
 *      251 user not local; will be forwarded to <email-address>
 *      550 user not found
 *      551 user not local; please try <email-address>
 *      553 user ambiguous
 * EXPN <username> <CRLF>
 *      250-real name <email-address>
 *      ...
 *      ...
 *      250 real name <email-address>
 *      550 access denied
 * HELP <string> <CRLF>
 * NOOP <CRLF>
 * QUIT <CRLF>
 * TURN <CRLF>
 *      502 refused
 * <reverse-path> => <path>
 * <forward-path> => <path>
 * <path> => "<" [ <a-d-l> ":" ] <mailbox> ">"
 * <a-d-l> => <at-domain> | <at-domain> "," <a-d-l>
 * <at-domain> => "@" <domain>
 * <domain> => <element> | <element> "." <domain>
 * <element> => <name> | "#" <number> | "[" <dotnum> "]"
 * <mailbox> => <local-part> "@" <domain>
 * <local-part> => <dot-string> | <quoted-string>
 * <name> => <a> <ldh-str> <let-dig>
 * <ldh-str> => <let-dig-hyp> | <let-dig-hyp> <ldh-str>
 * <let-dig> => <a> | <d>
 * <let-dig-hyp> => <a> | <d> | "-"
 * <dot-string> => <string> | <string> "." <dot-string>
 * <string> => <char> | <char> <string>
 * <quoted-string> => """ <qtext> """
 * <qtext> => "\" <x> | "\" <x> qtext | <q> | <q> <qtext>
 * <char> => <c> | "\" <x>
 * <dotnum> => <snum> "." <snum> "." <snum> "." <snum>
 * <number> => <d> | <d> <number>
 * <CRLF> => <CR> <LF>
 * <snum> => -- one, two or three digits representing a decimal in (0..255)
 * <a> => -- any of the 52 characters A through Z and a through z
 * <c> => -- any one of the 128 ASCII chars (excluding special or <SP>)
 * <d> => -- any one of the ten digits 0 through 9
 * <q> => -- any one of the 128 ASCII chars (excluding <CR>, <LF>, (") or (\)
 * <x> => -- any one of the 128 ASCII chars (no exceptions)
 * <special> => "<" | ">" | "(" | ")" | "[" | "]" | "\" | "." | "," | ";" |
 *              ":" | "@" | """ | -- control chars (0..31, 127)
 * replies:
 *   500 syntax error, command unrecognized (including line too long, ...)
 *   501 syntax error in parameters or arguments
 *   502 command not implemented
 *   504 command parameter not implemented
 *   211 system status ready/system help ready
 *   214 help message
 *   220 <domain> service ready
 *   221 <domain> service closing transmission channel
 *   421 <domain> service not available, closing transmission channel
 *   250 requested mail action okay, completed
 *   251 user not local; will forward to <forward-path>
 *   450 requested mail action not taken: mailbox unavailable (e.g. busy)
 *   550 requested action not taken: mailbox unavailable (not found, no access)
 *   451 requested action aborted: error in procession
 *   551 user not local; please try <forward-path>
 *   452 requested action not taken: insufficient system storage
 *   552 requested mail action aborted: exceeded storage allocation
 *   553 requested action not taken: mailbox name not allowed (syntax incorrect)
 *   354 start mail input; end with <CRLF>.<CRLF>
 *   554 transaction failed
 * When should each message code apply ("I"ntermediate, "S"uccess, "F"ailure
 * "E"rror, "P"reliminary):
 * CONNECTION ESTABLISHEMNT
 *   S: 220
 *   F: 421
 * HELO
 *   S: 250
 *   E: 500, 501, 504, 421
 * MAIL
 *   S: 250
 *   F: 552, 451, 452
 *   E: 500, 501, 421
 * RCPT
 *   S: 250, 251
 *   F: 550, 551, 552, 553, 450, 451, 452
 *   E: 500, 501, 503, 421
 * DATA
 *   I: 354 -> data -> S: 250
 *                     F: 552, 554, 451, 452
 *   F: 451, 554
 *   E: 500, 501, 503, 421
 * RSET
 *   S: 250
 *   E: 500, 501, 504, 421
 * SEND
 *   S: 250
 *   F: 552, 451, 452
 *   E: 500, 501, 502, 421
 * SOML
 *   S: 250
 *   F: 552, 451, 452
 *   E: 500, 501, 502, 421
 * SAML
 *   S: 250
 *   F: 552, 451, 452
 *   E: 500, 501, 502, 421
 * VRFY
 *   S: 250, 251
 *   F: 550, 551, 553
 *   E: 500, 501, 502, 504, 421
 * EXPN
 *   S: 250
 *   F: 550
 *   E: 500, 501, 502, 504, 421
 * HELP
 *   S: 211, 214
 *   E: 500, 501, 502, 504, 421
 * NOOP
 *   S: 250
 *   E: 500, 421
 * QUIT
 *   S: 221
 *   E: 500
 * TURN
 *   S: 250
 *   F: 502
 *   E: 500, 503
 * (Minimum implementation requirements for) Limits:
 *   user	- 64 chars
 *   domain	- 64 chars
 *   path	- 256 chars (including punctuation and element separators)
 *   commandline- 512 chars (including <CRLF>)
 *   reply line - 512 chars (including <CRLF>)
 *   text line  - 1000 chars (including <CRLF>, excluding the leading duplicated
 *                `.')
 *   recipients - 100 in one message
 *   >>: TO THE MAXIMUM EXTENT POSSIBLE,IMPLEMENTATION TECHNIQUES WHICH IMPOSE
 *       NO LIMITS ON THE LENGTH OF THESE OBJECTS SHOULD BE USED. :<<
 * Error due to exceeding these limits may be reported by using the reply codes:
 * (e.g.)
 *   500 Line too long.
 *   501 Path too long.
 *   552 Too many recipients.
 *   552 Too much mail data.
 * Extended SMTP commands
 *   EHLO <SP> <domain> <CRLF>
 *   BDAT <SP> chunk-size [ <SP> LAST ] <CRLF>
 *   xx
 */
struct smtp_cmds {
    SMTPcmd_t data,	/* start text of mail message */
	      ehlo,	/* request for smtp extensions */
	      etrn,	/* TURN SMTP-command in extended mode */
	      expn,	/* expand an address */
	      helo,	/* Introduce yourself */
	      help,	/* send a help text */
	      mail,	/* initiate a message with the sender's address */
	      noop,	/* do nothing */
	      quit,	/* terminate a SMTP session */
	      rcpt,	/* specify the first/next recipient for a message */
	      rset,	/* abort an initiated mail transfer */
	      saml,	/* Send And MaiL a message */
	      send,	/* send a message to a terminal */
	      soml,	/* Send Or MaiL a message */
	      turn,	/* tell the other side to send it's queued messages */
	      vrfy;	/* verify the deliverability of a message */
};

enum cmdtags { HELO_cmd, MAIL_cmd, RCPT_cmd, DATA_cmd, SEND_cmd, SOML_cmd,
	       SAML_cmd, RSET_cmd, VRFY_cmd, EXPN_cmd, HELP_cmd, NOOP_cmd,
	       QUIT_cmd, TURN_cmd, INVALID_cmd };
enum state { GREETING_state = 1, MAIL_state = 2, RCPT_state = 4,
	     DATA_state = 8, EOD_state = 16, QUIT_state = 32, SAVE_state = 64 };

static enum state state = GREETING_state;

#define RSET_state (GREETING_state|MAIL_state|RCPT_state|DATA_state)
#define VRFY_state (MAIL_state|RCPT_state|DATA_state)
#define BEGINNING_state (GREETING_state|MAIL_state)
static struct cmd_s {
    char *cmdname;
    enum cmdtags cmdtag;
    int valid_state;
    enum state next;
} cmds[] = {
    { "HELO", HELO_cmd, GREETING_state, MAIL_state },
    { "MAIL FROM:", MAIL_cmd,  BEGINNING_state, RCPT_state},
    { "RCPT TO:", RCPT_cmd, RCPT_state, DATA_state },
    { "DATA", DATA_cmd, RCPT_state, EOD_state },
    { "RSET", RSET_cmd, RSET_state, MAIL_state },
    { "QUIT", QUIT_cmd, RSET_state, QUIT_state },
    { "SEND FROM:", SEND_cmd, BEGINNING_state, MAIL_state },
    { "SOML FROM:", SOML_cmd, BEGINNING_state, MAIL_state },
    { "SAML FROM:", SAML_cmd, BEGINNING_state, MAIL_state },
    { "VRFY", VRFY_cmd, RSET_state, MAIL_state },
    { "EXPN", EXPN_cmd, RSET_state, MAIL_state },
    { "HELP", HELP_cmd, RSET_state, MAIL_state },
    { "NOOP", NOOP_cmd, RSET_state, MAIL_state },
    { "QUIT", QUIT_cmd, RSET_state, MAIL_state },
    { "TURN", TURN_cmd, RSET_state, MAIL_state },
    { NULL, INVALID_cmd, RSET_state, MAIL_state }
};

static char *getline (int *toolong)
{
    char line[1024], *p, *q;
    static char *res = NULL;
    static size_t ressz = 0;
    size_t llen, blen;
    int waseol;
    *toolong = 0;
    waseol = 0;
    llen = 0;
    while (!waseol) {
	if (llen == 0 && isNULL(fgets (line, 1024, stdin))) { return NULL; }
	waseol = cuteol (line);
	blen = strlen (line);
	if (llen + blen <= prefs.smtp_maxcmdlen) {
	    if (llen + blen + 1 > ressz) {
		ressz = llen + blen + 1;
		res = (isNULL(res) ? galloc (__FILE__, __LINE__, ressz) :
				     gresize (__FILE__, __LINE__, res, ressz));
	    }
	    strcpy (&res[llen], line); llen += blen;
	    continue;
	}
	*toolong = 1;
    }
    return res;
}

static void smtp_answer (int ecode, const char *fmt, ...)
{
    valist ap;
    char buf[1024], *p, *q, *r, *msg;
    size_t blen, rlen;
    msg = galloc (__FILE__, __LINE__, 1024);
    va_start (ap, fmt);
#if defined(__GLIBC__) && __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 1
    rlen = vsnprintf (msg, 1024, fmt, ap);
    if (rlen + 1 > 1024) {
	msg = gresize (__FILE__, __LINE__, msg, rlen + 1);
	vsnprintf (msg, rlen + 1, fmt, ap);
    }
#else
    blen = 1024;
    while (vsnprintf (msg, blen, fmt, ap) == -1) {
	blen += 1024;
	msg = gresize (__FILE__, __LINE__, msg, blen);
    }
#endif
    va_end (ap);
    q = msg;
    p = strchr (q, '\n');
    while (noNULL (p = strchr (q, '\n'))) {
	*p++ = '\0';
	fprintf (stdout, "%.3d%c%s\r\n", ecode, (*p ? '-' : ' '), q);
	q = p;
    }
    if (*q) { fprintf (stdout, "%.3d %s\r\n", q); }
    gfree (msg);
}

static int runcmd (char **args)
{
    char *cmdl, *sender = NULL, *recip = NULL, **recips = NULL, line[1024], *p;
    int toolong, ix, iy, nrecips = 0, tfc = 0, waseol, preveol;
    struct cmd_s *cmd;
    cmdl = getline (&toolong);
    if (toolong) {
	smtp_answer (500, "line too long");
	return NOOP_cmd;
    }
    for (ix = 0; noNULL (cmds[ix].cmdname); ++ix) {
	cmd = &cmds[ix];
	if (lccmp (cmdl, cmd->cmdname, strlen (cmd->cmdname))) { continue; }
	cmdl += strlen (cmd->cmdname);
	if (*cmdl && *cmdl[-1] != ':') { continue; }
	if (!(state & cmd->valid_state)) {
	    smtp_stateerror (cmd->valid_state); return 1;
	}
	switch (cmd->cmdtag) {
	    case HELO_cmd:
		smtp_answer (250, "Hello `%s'", cmdl);
		state = cmd->next;
		return 1;
	    case MAIL_cmd:
		sender = parse_addr (cmdl);
		if (check_sender (sender)) {
		    gfree (sender); sender = NULL;
		} else {
		    state = cmd->next;
		}
		return 1;
	    case RCPT_cmd:
		recip = parse_addr (cmdl);
		ifnull (strrchr (recip, '@')) {
		    char *p = gconc(__FILE__,__LINE__, recip,'@',domarg,NULL);
		    gfree (recip);
		    recip = p;
		}
		if (check_recip (recip)) {
		    gfree (recip); recip = NULL;
		} else {
		    ifnull (recips) {
			recips = (char **) galloc (__FILE__, __LINE__,
						   sizeof(char *));
			nrecips = 1;
		    } else {
			++nrecips;
			recips = (char **) gresize (__FILE__, __LINE__,
						    recips,
						    nrecips * sizeof (char *));
		    }
		    recips[nrecips - 1] = recip;
		}
		return 1;
	    case DATA_cmd:
		ifnull (recips) {
		    smtp_answer (503, "RCPT required!\n");
		    return 1;
		}
		tmpname = smtp_tmpfile (tfc++);
		ifnull (tf = sys_open (tmpname, "w", 0600)) {
		    smtp_answer (550, "local file system error - sorry\n... %s",
				      strerror (errno));
		    gfree (tmpname); return 1;
		}
		waseol = 1;
		while (fgets (line, 1024, stdin)) {
		    preveol = waseol;
		    waseol = cuteol (line);
		    p = line;
		    if (*p == '.' && preveol) { ++p; }
		    if (!*p && waseol) { break; }
		    fputs (p, tf);
		    if (waseol) { fputs ("\r\n", tf); }
		}
		smtp_answer (354, "message accepted for delivery");
		fclose (tf);
		for (iy = 0; iy < nrecips; ++iy) {
		    deliver_local (recips[iy], sender, recips[iy], tmpname);
		    gfree (recips[iy]);
		}
		unlink (tmpname); gfree (tmpname);
		gfree (sender); sender = NULL;
		gfree (recips); recips = NULL; nrecips = 0;
		state = cmd->next;
		return 1;
	    case QUIT_cmd:
		smtp_answer (221, "bye.");
		return 0;
	    case NOOP_cmd:
		smtp_answer (250, "OK");
		return 1;
	    case HELP_cmd:
		smtp_answer(214,
		    "This is VxDeliver version %s\n"
		    "The following commands are implemented:\n"
		    "  HELO [<hostname>] - begin a mail session\n"
		    "  MAIL FROM:sender  - specify the sender's address\n"
		    "  RCPT TO:recip     - specify recipient's address\n"
		    "  DATA              - begin the message\n"
		    "  QUIT              - end the SMTP session\n"
		    "  NOOP              - do nothing\n"
		    "  RSET              - reset the open transaction\n"
		    "\nYou can specify more than one recipient in a "
		    "single transaction.");
		return 1;
	    case RSET:
		smtp_answer (250, "Reset state");
		unlessnull (sender) { gfree (sender); }
		unlessnull (recips) {
		    for (iy = 0; iy < nrecips; ++iy) { gfree (recips[iy]); }
		    gfree (recips); recips = NULL; nrecips = 0;
		}
		state = cmd->next;
		return 1;
	    default:
		smtp_answer (550, "command not implemented (yet)");
		return 1;
	}
    }
    smtp_answer (500, "Command unrecognized: \"%s\"", cmdl);
    return 1;
}
